#include "Tools.h"
#include "matrix.h"
#include <sstream>
#include <fstream>
#include <iostream>
#include <iterator>
#include <iomanip>
#include <streambuf>
#include <string>
#include <list>
#include <algorithm>
#include <cstdlib>
#include <cmath>

using namespace std ;

/*******************/
/*                 */
/* Klasse AtomInfo */
/*                 */
/*******************/

class AtomInfo
{ private:
  string AtomTyp ;
  float X,Y,Z ;

  public:
  AtomInfo () : AtomTyp ("Unknown"), X(0.0), Y(0.0), Z(0.0) {} ;

  bool operator == (const AtomInfo AI) 
  { return ((AtomTyp == AI.AtomTyp) && (X == AI.X) && (Y == AI.Y) && (Z == AI.Z)) ; } ;
  bool operator < (const AtomInfo AI) { return (X < AI.X) ; } ;

  friend ostream& operator << (ostream& OutStream, const AtomInfo& AI) ;
  friend istream& operator >> (istream& InStream, AtomInfo& AI) ;
} ;

const char* AtomSymbole [87] =
{ "",
  "H" ,                                                                                "He",
  "Li","Be",                                                  "B" ,"C" ,"N" ,"O" ,"F" ,"Ne",
  "Na","Mg",                                                  "Al","Si","P" ,"S" ,"Cl","Ar",
  "K" ,"Ca","Sc","Ti","V" ,"Cr","Mn","Fe","Co","Ni","Cu","Zn","Ga","Ge","As","Se","Br","Kr",
  "Rb","Sr","Y" ,"Zr","Nb","Mo","Tc","Ru","Rh","Pd","Ag","Cd","In","Sn","Sb","Te","I" ,"Xe",
  "Cs","Ba","La",

  "Ce","Pr","Nd","Pm","Sm","Eu","Gd","Tb","Dy","Ho","Er","Tm","Yb","Lu",

                 "Hf","Ta","W" ,"Re","Os","Ir","Pt","Au","Hg","Tl","Pb","Bi","Po","At","Rn"
} ;

ostream& operator << (ostream& OutStream, const AtomInfo& AI) 
{ int AtomNr = 0 ;

  for (int I = 1; I < 87; ++I) if (AtomSymbole[I] == AI.AtomTyp) AtomNr = I ;

  OutStream << ' ' << AtomNr 
            << ' ' << setprecision (5) << 0.0 
      << ' ' << setprecision (5) << AI.X 
      << ' ' << setprecision (5) << AI.Y 
      << ' ' << setprecision (5) << AI.Z << endl ;
  return (OutStream) ;
}

istream& operator >> (istream& InStream, AtomInfo& AI) 
{ string Dummy ;
  float Bohr = float (0.529177249) ;

  InStream >> Dummy ;
  while ((Dummy != "ATOM") && InStream) InStream >> Dummy ;
  if (InStream) 
  { InStream >> Dummy >> AI.AtomTyp >> Dummy >> Dummy >> AI.X >> AI.Y >> AI.Z ;
    AI.X /= Bohr ;
    AI.Y /= Bohr ;
    AI.Z /= Bohr ;
  }

  return (InStream) ;
}

typedef list<AtomInfo> MolInfo ;

/******************/
/*                */
/* Klasse Minimax */
/*                */
/******************/

const float Gross = float (1e38) ;

class Minimax
{ private:
  float MinIso,MaxIso,MinAniso,MaxAniso ;
  float MinX,MaxX,MinY,MaxY,MinZ,MaxZ ;

  public:
  Minimax () 
  : MinIso (Gross), MaxIso (-Gross), MinAniso (Gross), MaxAniso (-Gross),
    MinX (Gross), MaxX (-Gross), MinY (Gross), MaxY (-Gross), MinZ (Gross), MaxZ (-Gross)
  {
  }
  
  float NormOrt (int Ebene,float Wert) ;
  float RealOrt (int Ebene,float Wert) ;
  float NormIso (float Wert) ;
  float NormAniso (float Wert) ;
  float DiffX () { return (MaxX-MinX) ; } ;
  float DiffY () { return (MaxY-MinY) ; } ;
  float DiffZ () { return (MaxZ-MinZ) ; } ;
  void CheckGrenzen (const Minimax& MM)
  { CheckGrenzen (MM.MinIso,MM.MaxIso,MM.MinAniso,MM.MaxAniso,MM.MinX,MM.MaxX,MM.MinY,MM.MaxY,MM.MinZ,MM.MaxZ) ;
  }
  void CheckGrenzen (float aMinIso,float aMaxIso,float aMinAniso,float aMaxAniso,float aMinX,float aMaxX,float aMinY,float aMaxY,float aMinZ,float aMaxZ) ;
  friend ostream& operator << (ostream& OutStream, const Minimax& MM) ;
  friend istream& operator >> (istream& InStream, Minimax& MM) ;
} ;

void Minimax::CheckGrenzen (float aMinIso,float aMaxIso,float aMinAniso,float aMaxAniso,float aMinX,float aMaxX,float aMinY,float aMaxY,float aMinZ,float aMaxZ)
{ if (MinIso > aMinIso) MinIso = aMinIso ;
  if (MaxIso < aMaxIso) MaxIso = aMaxIso ;
  if (MinAniso > aMinAniso) MinAniso = aMinAniso ;
  if (MaxAniso < aMaxAniso) MaxAniso = aMaxAniso ;
  if (MinX > aMinX) MinX = aMinX ;
  if (MaxX < aMaxX) MaxX = aMaxX ;
  if (MinY > aMinY) MinY = aMinY ;
  if (MaxY < aMaxY) MaxY = aMaxY ;
  if (MinZ > aMinZ) MinZ = aMinZ ;
  if (MaxZ < aMaxZ) MaxZ = aMaxZ ;
}

float Minimax::NormIso (float Wert) 
{ if (MaxIso == MinIso) return (0) ; else 
  { float Ergebnis = (Wert-MinIso)/(MaxIso-MinIso) ;
    if (Ergebnis < 0) Ergebnis = 0 ;
    if (Ergebnis > 1) Ergebnis = 1 ;
    return (Ergebnis) ;
  }
}

float Minimax::NormAniso (float Wert) 
{ if (MaxAniso == MinAniso) return (0) ; else 
  { float Ergebnis = ((Wert)-(MinAniso))/(MaxAniso-MinAniso) ;
    if (Ergebnis < 0) Ergebnis = 0 ;
    if (Ergebnis > 1) Ergebnis = 1 ;
    return (Ergebnis) ;
    //return ((Wert-MinAniso)/(MaxAniso-MinAniso)) ;
  }
}

float Minimax::NormOrt (int Ebene,float Wert) 
{ float Ergebnis ;

  switch (Ebene)
  { case 0: { Ergebnis = (Wert-MinX)/(MaxX-MinX) ; break ; } ;
    case 1: { Ergebnis = (Wert-MinY)/(MaxY-MinY) ; break ; } ;
    case 2: { Ergebnis = (Wert-MinZ)/(MaxZ-MinZ) ; break ; } ;
  }
  if (Ergebnis < 0) Ergebnis = 0 ;
  if (Ergebnis > 1) Ergebnis = 1 ;
  return (Ergebnis) ;  
}

float Minimax::RealOrt (int Ebene,float Wert) 
{ float Ergebnis ;
  
  if (Wert < 0) Wert = 0 ;
  if (Wert > 1) Wert = 1 ;

  switch (Ebene)
  { case 0: { Ergebnis = Wert*(MaxX-MinX)+MinX ; break ; } ;
    case 1: { Ergebnis = Wert*(MaxY-MinY)+MinY ; break ; } ;
    case 2: { Ergebnis = Wert*(MaxZ-MinZ)+MinZ ; break ; } ;
  }
  return (Ergebnis) ;
}

ostream& operator << (ostream& OutStream, const Minimax& MM) 
{ OutStream << MM.MinIso << " <= Isotropie <= " << MM.MaxIso << endl
            << MM.MinAniso << " <= Anisotropie <= " << MM.MaxAniso << endl
            << MM.MinX << " <= X <= " << MM.MaxX << endl
            << MM.MinY << " <= Y <= " << MM.MaxY << endl
            << MM.MinZ << " <= Z <= " << MM.MaxZ << endl ;
  return (OutStream) ;
}

istream& operator >> (istream& InStream, Minimax& MM) 
{ string Dummy ;

  InStream >> MM.MinIso >> Dummy >> Dummy >> Dummy >> MM.MaxIso
           >> MM.MinAniso >> Dummy >> Dummy >> Dummy >> MM.MaxAniso
           >> MM.MinX >> Dummy >> Dummy >> Dummy >> MM.MaxX
           >> MM.MinY >> Dummy >> Dummy >> Dummy >> MM.MaxY
           >> MM.MinZ >> Dummy >> Dummy >> Dummy >> MM.MaxZ ;
  return (InStream) ;
}

/*******************/
/*                 */
/* Klasse KartNetz */
/*                 */
/*******************/

class KartNetz
{ private:
  Minimax Grenzen ;
  float* Daten ;
  int AnzahlGitterpunkte ;
  Matrix<int> Max ;

  public:
  KartNetz () : AnzahlGitterpunkte (0), Daten (0), Max (Null,1,3) {} ;
  KartNetz (istream& InStream) ;
  ~KartNetz () ;

  void ParameterAusgabe (ostream& OutStream) const ;
  void SchreibePov (ostream& OutStream,MolInfo Molekuel) ;

  float* At (int x,int y,int z) ;
  inline float RealWert (int Koordinate, int Wert) ;
} ;

KartNetz::KartNetz (istream& InStream)
: AnzahlGitterpunkte (0), Max (Null,1,3)
{ string Dummy ;

  InStream >> Dummy >> AnzahlGitterpunkte >> Dummy >> *Max[0] >> Dummy >> *Max[1] >> Dummy >> *Max[2] 
           >> Dummy >> Grenzen ;
  InStream.ignore (1) ;

//  cout << "Streampos: " << InStream.tellg () << endl ;

  assert (AnzahlGitterpunkte == Max.x()*Max.y()*Max.z()) ;
  Daten = new float [AnzahlGitterpunkte*9] ;
  InStream.read ((char *)Daten,sizeof (float)*9*AnzahlGitterpunkte) ;
}

KartNetz::~KartNetz ()
{ delete [] Daten ;
}

inline float KartNetz::RealWert (int Koordinate, int Wert)
{ return (Grenzen.RealOrt (Koordinate,float (Wert) / (*Max[Koordinate]-1))) ;
}

float* KartNetz::At (int x,int y,int z)
{ return (&Daten[((z*Max.y()+y)*Max.x()+x)*9]) ;
}

void KartNetz::ParameterAusgabe (ostream& OutStream) const
{ OutStream << AnzahlGitterpunkte << " = " << *Max[0] << " * " << *Max[1] << " * " << *Max[2] << endl
            << "Grenzen:" << endl << Grenzen << endl ;
}

void KartNetz::SchreibePov (ostream& OutStream,MolInfo Molekuel)
{ float* Anisotropie = new float [AnzahlGitterpunkte] ;

  // ber Daten iterieren
  float* AnisotropieAkt = Anisotropie ;
  { for (int Z = 0; Z < Max.z(); ++Z)
    { for (int Y = 0; Y < Max.y(); ++Y)
      { for (int X = 0; X < Max.x(); ++X)
        { // Anisotropie berechnen
          Matrix<float> CurMatrix (At(X,Y,Z),3,3) ;
          *AnisotropieAkt++ = CurMatrix.Anisotropie () ;
        }
      }
    }
  }

  cerr << "Cube-file wird geschrieben. Bitte warten." << endl ;

  float StepX = Grenzen.DiffX () / Max.x() ;
  float StepY = Grenzen.DiffY () / Max.y() ;
  float StepZ = Grenzen.DiffZ () / Max.z() ;

  // Cube-file-header ausgeben
  OutStream << " (empty)" << endl
            << " Anisotropy of the Induced Current Density" << endl
            << ' ' << Molekuel.size () << "  " << RealWert (0,0)+StepX/2 << "  " << RealWert (1,0)+StepY/2 << "  " << RealWert (2,0)+StepZ/2 << endl 
            << ' ' << Max.x() << "  " << StepX << "  0.00000  0.00000" << endl
            << ' ' << Max.y() << "  0.00000  " << StepY << "  0.00000" << endl
            << ' ' << Max.z() << "  0.00000  0.00000  " << StepZ << endl ;

  // Moleklinformation ausgeben
  OutStream.setf (ios::showpoint) ;
  ostream_iterator<AtomInfo> MolPrinter (OutStream) ;
  copy (Molekuel.begin(),Molekuel.end(),MolPrinter) ;
  
  // Alle Punkte ausgeben
  OutStream.setf (ios::showpoint | ios::scientific | ios::uppercase) ;
  AnisotropieAkt = Anisotropie ;
  { for (int X = 0; X < Max.x(); ++X)
    { for (int Y = 0; Y < Max.y(); ++Y)
      { for (int Z = 0; Z < Max.z(); ++Z)
        { OutStream << setprecision (5) << "  " << Anisotropie[((Z*Max.y())+Y)*Max.x()+X] << endl ;
        }
      }
    }
  }

  // Speicher freigeben
  delete [] Anisotropie ;
}

/**********************/
/*                    */
/* globale Funktionen */
/*                    */
/**********************/

void Hilfe () 
{ cout << endl 
       << "AICD-cube - Erzeugung eines cube-files aus dem Output" << endl
       << "            von AICD-remap" << endl
       << endl
       << "Dieses Programm liest die Stromdichtedaten aus dem" << endl
       << "output-file von AICD-remap und die Strukturdaten des" << endl
       << "Molekuels aus einem pdb-file und produziert daraus ein" << endl
       << "cube-file der Anisotropie der Stromdichtetensoren. Das" << endl
       << "cube-file kann z.B. von GaussView gelesen und zur Anzeige" << endl
       << "der Isooberflaeche verwendet werden." << endl
       << endl
       << "Die input-files muessen in zwei Listen angegeben werden." << endl
       << "Zunaechst folgen die output-files von AICD-remap und" << endl
       << "schliesslich die zugehoerigen pdb-files in derselben" << endl
       << "Reihenfolge." << endl
       << endl
       << "Die output-files erhalten den Namen des output-files von" << endl
       << "AICD-remap mit einem angehaengten \".cub\"." << endl
       << endl
       << "Aufruf: AICD-cube  input-file(s)  input-file(s) (.pdb-files)" << endl
       << "Output: input-file.cub" << endl 
       << endl ;  
} 

/*****************/
/*               */
/* Hauptprogramm */
/*               */
/*****************/

int main (int argc, char* argv [])
{
	if ((argc < 3) || ((argc & 1) == 0)) Hilfe () ; else
  for (int ArgIndex = 0; ArgIndex < (argc >> 1); ++ArgIndex)
  { char *Inputfilename, *PDBFilename ;

    // Parameter einlesen    
    Inputfilename = argv[ArgIndex+1] ;
    PDBFilename = argv[ArgIndex+1+(argc >> 1)] ;

    cout << endl << ArgIndex+1 << ". Dateienpaar: " << Inputfilename << " und " << PDBFilename << endl ;

    // Einlesen des PDB-Files
    ifstream PDBFile (PDBFilename) ;
    MolInfo Molekuel ;
    while (PDBFile)
    { AtomInfo Temp ;
      PDBFile >> Temp ;
      if (PDBFile) Molekuel.push_back (Temp) ;
    }

    // Moleklinformation ausgeben
    cout << endl << "Moleklinformationen:" << endl ;
    ostream_iterator<AtomInfo> MolPrinter (cout) ;
    copy (Molekuel.begin(),Molekuel.end(),MolPrinter) ;
    cout << endl ;

    // Einlesen der Daten aus dem input-file
    ifstream Inputfile (Inputfilename,ios::binary) ;
    KartNetz globalesNetz (Inputfile) ;
    globalesNetz.ParameterAusgabe (cout) ;

    // cub-Datei schreiben
    ofstream CubeDatei ((string(Inputfilename)+".cub").c_str(),ios::binary) ;
    globalesNetz.SchreibePov (CubeDatei,Molekuel) ;
  }
  return 0;
}
