//#define IRIX
#ifndef matrix_h
#define matrix_h
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <cassert>
#include <cmath>

using namespace std;

#ifndef LINUX
//const double PI = double (3.1415926) ;
#endif

enum MatrixTyp { Null, Einheit, Undefiniert } ;

// Forward references for using templated friends
// ----------------------------------------------
template <typename T> class Matrix;
template <typename T> ostream& operator << (ostream& OutStream, const Matrix<T>& M) ;
template <typename T> istream& operator >> (istream& InStream, Matrix<T>& M) ;
template <typename T> Matrix<T> operator / (const Matrix<T>& Dividend, T Divisor) ;
//template <typename T> Matrix<T> operator * (const Matrix<T>& Faktor1, T Faktor2) ;
//template <typename T> Matrix<T> operator * (T Faktor2, const Matrix<T>& Faktor1) ;

// Start of real definitions
// --------------------------
template <typename T>
class Matrix
#ifdef trace_on
: virtual public trace
#endif
{ int Spalten, Zeilen ;
  T* Daten ;
  T* EndDaten ;

  void DatenAllozierenKopieren (T* daten) ;
  void DatenAllozierenInitialisieren (MatrixTyp Typ) ;

  public:
  Matrix () ;
  Matrix (const Matrix<T>&) ;
  Matrix (T* daten, int spalten, int zeilen) ;
  Matrix (MatrixTyp Typ, int spalten, int zeilen) ;
  Matrix (T* daten, int spalten = 3) ;
  Matrix (MatrixTyp Typ, int spalten = 3) ;
  virtual ~Matrix () ;

  bool       operator ==   (const Matrix<T>&) const ;
  inline bool       operator !=   (const Matrix<T>& m) const {return !(*this==m);};
  Matrix<T>& operator =    (const Matrix<T>&) ;
  Matrix<T>& operator +=   (const Matrix<T>& Summand) ;
  Matrix<T>  operator +    (const Matrix<T>& Summand) const ;
  Matrix<T>  operator -    () const ;
  Matrix<T>& operator -=   (const Matrix<T>& Summand) ;
  Matrix<T>  operator -    (const Matrix<T>& Summand) const ;
  T          SkalarProdukt (const Matrix<T>& Faktor) const ;
  Matrix<T>  KreuzProdukt  (const Matrix<T>& Faktor) const ;
  Matrix<T>  operator *    (const Matrix<T>& Faktor) const ;
  Matrix<T>& operator *=   (const Matrix<T>& Faktor) ;

  Matrix<T>& operator *=   (T Faktor) ;
  Matrix<T>& operator /=   (T Divisor) ;

  T* operator [] (int ZeilenIndex) const ;

  T Isotropie () const ;
  T Anisotropie () const ;
  T FullAnisotropy () const ;
  T AntisymmetricAnisotropy () const ;
  void write (ostream& OutStream) const ;
  T x () const ;
  T y () const ;
  T z () const ;
  T Laenge () const ;
  Matrix<T> ZeilenSpaltenVertauschen () ;


#ifdef IRIX
  friend ostream& operator << <> (ostream& OutStream, const Matrix<T>& M) ;
  friend istream& operator >> <> (istream& InStream, Matrix<T>& M) ;
//  friend Matrix<T> operator * <> (const Matrix<T>& Faktor1, T Faktor2) ;
//  friend Matrix<T> operator * <> (T Faktor2, const Matrix<T>& Faktor1) ;
  friend Matrix<T> operator / <> (const Matrix<T>& Dividend, T Divisor) ;
#else
  friend ostream& operator << <> (ostream& OutStream, const Matrix<T>& M) ;
  friend istream& operator >> <> (istream& InStream, Matrix<T>& M) ;
  friend Matrix<T> operator / <> (const Matrix<T>& Dividend, T Divisor) ;
//  friend Matrix<T> operator * <> (const Matrix<T>& Faktor1, T Faktor2) ;
//  friend Matrix<T> operator * <> (T Faktor2, const Matrix<T>& Faktor1) ;
#endif
} ;


/*******************/
/*                 */
/* Hilfsfunktionen */
/*                 */
/*******************/

template <typename T>
inline T sqr (T Argument)
{ return (Argument*Argument) ;
}

/****************************************/
/*                                      */
/* Klasse Matrix<T>: private Funktionen */
/*                                      */
/****************************************/

template <typename T>
void Matrix<T>::DatenAllozierenKopieren (T* daten)
{ int Anzahl = Spalten*Zeilen ;
  Daten = new T [Anzahl] ;
  EndDaten = Daten + Anzahl ;
  if (daten) std::memcpy (Daten, daten, sizeof (T)*Anzahl) ;
}

template <typename T>
void Matrix<T>::DatenAllozierenInitialisieren (MatrixTyp Typ)
{ DatenAllozierenKopieren (NULL) ;

  if (Typ != Undefiniert)
     for (T* TempDaten = Daten; TempDaten < EndDaten; TempDaten++)
         *TempDaten = 0 ;

  if (Typ == Einheit)
     for (T* TempDaten = Daten; TempDaten < EndDaten; TempDaten += Spalten+1)
         *TempDaten = 1 ;
}

/************************************************/
/*                                              */
/* Klasse Matrix<T>: Konstruktoren/Destruktoren */
/*                                              */
/************************************************/

template <typename T>
Matrix<T>::Matrix ()
: Spalten (0), Zeilen (0), Daten (NULL), EndDaten (NULL)
#ifdef trace_on 
, trace ("Matrix")
#endif
{
}

template <typename T>
Matrix<T>::Matrix (const Matrix<T>& Original)
: Spalten (Original.Spalten), Zeilen (Original.Zeilen)
#ifdef trace_on 
, trace ("Matrix")
#endif
{ DatenAllozierenKopieren (Original.Daten) ;
}

template <typename T>
Matrix<T>::Matrix (T* daten, int spalten, int zeilen)
: Spalten (spalten), Zeilen (zeilen)
#ifdef trace_on 
, trace ("Matrix")
#endif
{ DatenAllozierenKopieren (daten) ;
}

template <typename T>
Matrix<T>::Matrix (T* daten, int spalten)
: Spalten (spalten), Zeilen (spalten)
#ifdef trace_on 
, trace ("Matrix")
#endif
{ DatenAllozierenKopieren (daten) ;
}

template <typename T>
Matrix<T>::Matrix (MatrixTyp Typ, int spalten, int zeilen)
: Spalten (spalten), Zeilen (zeilen)
#ifdef trace_on 
, trace ("Matrix")
#endif
{ assert ((Typ != Einheit) || (zeilen == spalten)) ;
  DatenAllozierenInitialisieren (Typ) ;
}

template <typename T>
Matrix<T>::Matrix (MatrixTyp Typ, int spalten)
: Spalten (spalten), Zeilen (spalten)
#ifdef trace_on
, trace ("Matrix")
#endif
{ DatenAllozierenInitialisieren (Typ) ;
}

template <typename T>
Matrix<T>::~Matrix ()
{ delete [] Daten ;
}

/********************************/
/*                              */
/* Klasse Matrix<T>: Operatoren */
/*                              */
/********************************/

template <typename T>
bool Matrix<T>::operator == (const Matrix<T>& Vergleich) const 
{ T* TempDaten1 = Daten ;
  T* TempDaten2 = Vergleich.Daten ;
  bool Ergebnis = ((Zeilen == Vergleich.Zeilen) && (Spalten == Vergleich.Spalten)) ;

  while (Ergebnis && (TempDaten1 < EndDaten)) Ergebnis = (TempDaten1++ == TempDaten2++) ;
  return (Ergebnis) ;
}

template <typename T>
Matrix<T>& Matrix<T>::operator = (const Matrix<T>& Original)
{ if (this != &Original)
  { delete [] Daten ;
    Spalten = Original.Spalten ;
    Zeilen = Original.Zeilen ;

    DatenAllozierenKopieren (Original.Daten) ;
  }
  return (*this) ;
}

template <typename T>
Matrix<T>& Matrix<T>::operator += (const Matrix<T>& Summand)
{ assert ((Spalten == Summand.Spalten) && (Zeilen == Summand.Zeilen)) ;

  T* SummandDaten = Summand.Daten ;

  for (T* TempDaten = Daten; TempDaten < EndDaten; *TempDaten++ += *SummandDaten++) ;
  return (*this) ;
}

template <typename T>
Matrix<T> Matrix<T>::operator + (const Matrix<T>& Summand) const
{ Matrix<T> TempMatrix = *this ;
  return (TempMatrix += Summand) ;
}

template <typename T>
Matrix<T> Matrix<T>::operator - () const
{ Matrix<T> TempMatrix = *this ;
  for (T* TempDaten = TempMatrix.Daten; TempDaten < TempMatrix.EndDaten; *TempDaten = -*TempDaten++) ;
  return (TempMatrix) ;
}

template <typename T>
inline Matrix<T>& Matrix<T>::operator -= (const Matrix<T>& Summand)
{ return (*this += -Summand) ;
}

template <typename T>
Matrix<T> Matrix<T>::operator - (const Matrix<T>& Summand) const
{ Matrix<T> TempMatrix = *this ;
  return (TempMatrix -= Summand) ;
}

template <typename T>
T Matrix<T>::SkalarProdukt (const Matrix<T>& Faktor) const 
{ // Skalarprodukt
  assert ((Spalten == 1) && (Faktor.Spalten == 1) && (Zeilen == Faktor.Zeilen)) ;
  double Ergebnis = 0.0 ;
  
  T* TempDaten1 = Daten ;
  T* TempDaten2 = Faktor.Daten ;
  for (int zeilen = 0; zeilen < Zeilen; ++zeilen) Ergebnis += *TempDaten1++ * *TempDaten2++ ;

  return (Ergebnis) ;
}

template <typename T>
Matrix<T> Matrix<T>::KreuzProdukt (const Matrix<T>& Faktor) const 
{ // Skalarprodukt
  assert ((Spalten == 1) && (Faktor.Spalten == 1) && (Zeilen == Faktor.Zeilen) && (Zeilen == 3)) ;
  Matrix<T> Ergebnis (Null,1,3) ;
  
  *Ergebnis [0] = (*(*this) [1]) * (*Faktor [2]) - (*(*this) [2]) * (*Faktor [1]) ;
  *Ergebnis [1] = (*(*this) [2]) * (*Faktor [0]) - (*(*this) [0]) * (*Faktor [2]) ;
  *Ergebnis [2] = (*(*this) [0]) * (*Faktor [1]) - (*(*this) [1]) * (*Faktor [0]) ;

  return (Ergebnis) ;
}

template <typename T>
Matrix<T> Matrix<T>::operator * (const Matrix<T>& Faktor) const
{ assert (Spalten == Faktor.Zeilen) ;

  Matrix<T> TempMatrix (Null, Faktor.Spalten, Zeilen) ;
  T* TempDaten = TempMatrix.Daten ;
  for (int zeilen = 0; zeilen < TempMatrix.Zeilen; zeilen++)
  { for (int spalten = 0; spalten < TempMatrix.Spalten; spalten++)
    { *TempDaten = 0 ;
      for (int Index = 0; Index < Spalten; Index++)
      { *TempDaten += (*this)[zeilen][Index] * Faktor[Index][spalten] ;
      }
      TempDaten++ ;
    }
  }

  return (TempMatrix) ;
}

template <typename T>
inline Matrix<T>& Matrix<T>::operator *= (const Matrix<T>& Faktor)
{ return (*this = *this * Faktor) ;
}

template <typename T>
Matrix<T>& Matrix<T>::operator *= (T Faktor)
{ for (T* TempDaten = Daten; TempDaten < EndDaten; *TempDaten++ *= Faktor) ;
  return (*this) ;
}

template <typename T>
inline Matrix<T>& Matrix<T>::operator /= (T Divisor)
{ return (*this *= 1/Divisor) ;
}

template <typename T>
inline T* Matrix<T>::operator [] (int ZeilenIndex) const
{ assert ((ZeilenIndex >= 0) && (ZeilenIndex < Zeilen)) ;

  return (Daten+ZeilenIndex*Spalten) ;
}

template <typename T>
Matrix<T> Matrix<T>::ZeilenSpaltenVertauschen () 
{ Matrix<T> Temp (Null,Zeilen,Spalten) ;

  for (int zeilen = 0; zeilen < Zeilen; zeilen++)
  { for (int spalten = 0; spalten < Spalten; spalten++)
    { Temp[spalten][zeilen] = (*this)[zeilen][spalten] ;
    }
  }

  return (Temp) ;
}

/**************************************************/
/*                                                */
/* Klasse Matrix<T>: friend Operatoren/Funktionen */
/*                                                */
/**************************************************/

template <typename T>
Matrix<T> operator * (const Matrix<T>& Faktor1, T Faktor2)
{ Matrix<T> TempMatrix (Faktor1) ;
  return (TempMatrix *= Faktor2) ;
}

template <typename T>
inline Matrix<T> operator * (T Faktor2, const Matrix<T>& Faktor1)
{ return (Faktor1*Faktor2) ;
}

template <typename T>
Matrix<T> operator / (const Matrix<T>& Dividend, T Divisor)
{ Matrix<T> TempMatrix (Dividend) ;
  return (TempMatrix /= Divisor) ;
}

template <typename T>
ostream& operator << (ostream& OutStream, const Matrix<T>& M)
{
//for (T* TempDaten = M.Daten; TempDaten < M.EndDaten; OutStream << *TempDaten++ << "  ") ;
  for (int zeilen = 0; zeilen < M.Zeilen; zeilen++)
  { for (int spalten = 0; spalten < M.Spalten; spalten++)
    { OutStream << M[zeilen][spalten] << "  " ;
    }
    OutStream << endl ;
  }
  return (OutStream) ;
}

template <typename T>
istream& operator >> (istream& InStream, Matrix<T>& M)
{
//for (T* TempDaten = M.Daten; TempDaten < M.EndDaten; OutStream << *TempDaten++ << "  ") ;
  for (int zeilen = 0; zeilen < M.Zeilen; zeilen++)
  { for (int spalten = 0; spalten < M.Spalten; spalten++)
    { InStream >> M[zeilen][spalten] ;
    }
  }
  return (InStream) ;
}

// specialize the input operator to avoid errors when
// reading floating point numbers of a higher precision. eg.
//    Matrix<float> >> double
template <>
istream& operator  >>  (istream& InStream, Matrix<float>& M)
{
//for (T* TempDaten = M.Daten; TempDaten < M.EndDaten; OutStream << *TempDaten++ << "  ") ;
  double d;
  for (int zeilen = 0; zeilen < M.Zeilen; zeilen++)
  { for (int spalten = 0; spalten < M.Spalten; spalten++)
    {
      InStream >> d;
      M[zeilen][spalten] = d;
    }
  }
  return (InStream);
}


template <typename T>
Matrix<T> RotationsMatrix (int i_Koordinate, T T_Winkel)
{
  Matrix<T> Abbildung (Einheit,3,3) ;
  int i_klein = (i_Koordinate==0)? 1 : 0 ;
  int i_gross = (i_Koordinate==2)? 1 : 2 ;

  Abbildung[i_klein][i_klein] = cos (T_Winkel) ;
  Abbildung[i_gross][i_gross] = Abbildung[i_klein][i_klein] ;
  Abbildung[i_klein][i_gross] = - sin (T_Winkel) ;
  Abbildung[i_gross][i_klein] = - Abbildung[i_klein][i_gross] ;

  return (Abbildung) ;
}


/********************************************/
/*                                          */
/* Klasse Matrix<T>: ffentliche Funktionen */
/*                                          */
/********************************************/

template <typename T>
T Matrix<T>::Isotropie () const 
{ assert (Spalten == Zeilen) ;

  T Ergebnis = 0 ;

  for (int Index = 0; Index < Spalten; Index++) Ergebnis += (*this)[Index][Index] ;
  return (Ergebnis / Spalten) ;  
}

template <typename T>
T Matrix<T>::Anisotropie () const 
{ assert ((Spalten == 3) && (Zeilen == 3)) ;

  return (sqrt
            ( (0.5) *
               (  sqr( (*this)[0][0]-(*this)[1][1] )
                + sqr( (*this)[1][1]-(*this)[2][2] )
                + sqr( (*this)[2][2]-(*this)[0][0] )
                + 1.5* (  sqr( (*this)[0][1]+(*this)[1][0] )
                        + sqr( (*this)[1][2]+(*this)[2][1] )
                        + sqr( (*this)[0][2]+(*this)[2][0] )
                       )
               )
            )
         ) ;
}
  
template <class T>
T Matrix<T>::FullAnisotropy () const 
{ assert ((Spalten == 3) && (Zeilen == 3)) ;

  return (sqrt
            ( abs
               ( (0.5) *
                  (  sqr( (*this)[0][0]-(*this)[1][1] )
                   + sqr( (*this)[1][1]-(*this)[2][2] )
                   + sqr( (*this)[2][2]-(*this)[0][0] )
                   + 6* (  (*this)[0][1]*(*this)[1][0]
                         + (*this)[1][2]*(*this)[2][1]
                         + (*this)[0][2]*(*this)[2][0]
                        )
                  )
               )
            )
         ) ;
}
  
template <class T>
T Matrix<T>::AntisymmetricAnisotropy () const 
{ assert ((Spalten == 3) && (Zeilen == 3)) ;

  return (sqrt
            ( abs
               ( (-0.75) *
                  (  sqr( (*this)[0][1]-(*this)[1][0] )
                   + sqr( (*this)[1][2]-(*this)[2][1] )
                   + sqr( (*this)[0][2]-(*this)[2][0] )
                  )
               )
            )
         ) ;
}
  
template <typename T>
inline void Matrix<T>::write (ostream& OutStream) const
{ if (Daten) OutStream.write ((const char *) Daten,sizeof (T)*Spalten*Zeilen) ;
}

template <typename T>
T Matrix<T>::x () const { assert ((Spalten == 1) && (Zeilen == 3)) ; return (Daten [0]) ; }

template <typename T>
T Matrix<T>::y () const { assert ((Spalten == 1) && (Zeilen == 3)) ; return (Daten [1]) ; }

template <typename T>
T Matrix<T>::z () const { assert ((Spalten == 1) && (Zeilen == 3)) ; return (Daten [2]) ; }

template <typename T>
T Matrix<T>::Laenge () const 
{ assert ((Spalten == 1) && (Zeilen == 3)) ; 
  return (T(sqrt(x()*x()+y()*y()+z()*z()))) ;
}




template <typename T>
Matrix<T> RotationsMatrix (int i_Koordinate, T T_Winkel);


/*******************/
/*                 */
/* Hilfsfunktionen */
/*                 */
/*******************/

template <typename T>
inline T sqr (T Argument);

#endif      
